#include <unistd.h>
#include <cstring>
#include <vector>
#include <sstream>
#include <sys/statvfs.h>
#include <sys/stat.h>
#include <netdb.h>
#include <signal.h>
#include "libalc.h"
#include "readline.h"
#include "alc_constants.h"
#include "aes_crypto.h"
#include "base64.h"
#include "alc_config.h"
#include "uuid.h"

using namespace std;

int net_read(LLS* config, char* buffer, size_t num_bytes) {
    // TODO: handle if connection has been broken
//    LOGD("Reading string of %d bytes\n", num_bytes);
    int ret = readLine(config->comm_socket, buffer, num_bytes);
//    LOGD("Returned result: %d\n", ret);
//    LOGD("Returned string: %s\n", buffer);
    if (ret == -1) {
        // The connection has been severed, die!
        kill(getpid(), SIGTERM);
    }
    return ret;
}

int net_write(LLS* config, char* buffer, size_t num_bytes) {
    // TODO: handle if connection has been broken
//    LOGD("Writing string: %s", buffer);
//    LOGD("String length: %d bytes\n", num_bytes);
//    LOGD("Socket: %d\n", config->comm_socket);
    int ret = write(config->comm_socket, buffer, num_bytes);
    if (ret == -1) {
        kill(getpid(), SIGTERM);
    }
//    LOGD("Returned result: %d\n", ret);
    return ret;
}

int net_write(LLS* config, const char* message) {
    return net_write(config, (char*) message, strlen(message));
}

int net_write(LLS* config, string buffer) {
    return net_write(config, buffer.c_str());
}

int send_command(LLS* config, vector<string> commands) {
//    LOGD("Sending command\n");
    stringstream buffer;
    for(string command : commands) {
        buffer << command;
        buffer << WORD_SEPARATOR;
    }
    // encrypt before endline
    if (config->encrypted) {
//        LOGD("ENCRYPTING\n");
        string buffer_string = buffer.str();
//        LOGD("BUFFER STRING: %s\n", buffer_string.c_str());
        string encrypted_string = encrypt_base64((unsigned char *) buffer_string.c_str(), buffer_string.length(),
                                                 config->encryption_key_bytes);
        string to_send = encrypted_string + "\n";
//        LOGD("Processed command string: %s", to_send.c_str());
        return net_write(config, to_send);
    } else {
//        LOGD("NOT ENCRYPTING");
        buffer << "\n";
        string buffer_string = buffer.str();
//        LOGD("Processed command string: %s", buffer_string.c_str());
        return net_write(config, buffer_string);
    }
}

char* global_buffer = NULL;
unsigned char* global_newbuffer = NULL;

void read_command(LLS* config, vector<string> *target) {
    //char buffer[MAX_COMMAND_LENGTH];
    if (global_buffer == NULL) {
        global_buffer = (char *) malloc(MAX_COMMAND_LENGTH);
    }
    if (global_newbuffer == NULL) {
        global_newbuffer = (unsigned char *) malloc(MAX_COMMAND_LENGTH);
    }
    char* buffer = global_buffer;
    int ret = net_read(config, buffer, MAX_COMMAND_LENGTH);
    if (ret == -1) {
        perror(READ_COMMAND_FAILED_MSG);
    }
    buffer[ret-1] = '\0';
    // last received character should be newline, we don't want to return newline
    stringstream bstream;
    if (config->encrypted) {
//        LOGD("Encrypted stream!\n");
        string ciphertext(buffer, ret-1);
        unsigned char* newbuffer = global_newbuffer;
        int end = decrypt_base64(ciphertext, config->encryption_key_bytes, newbuffer);
//        LOGD("Decrypted successfully! Decrypted length: %d\n", end);
        newbuffer[end] = 0;
        bstream << newbuffer;
    } else {
//        LOGD("Not Encrypted stream!\n");
        bstream << buffer;
    }
    string tempstring;
    while (bstream >> tempstring) {
        target->push_back(tempstring);
    }
}

vector<string> * read_command(LLS* config) {
    vector<string> *result = new vector<string>();
    read_command(config, result);
//    LOGD("Read command successful!\n");
    return result;
}

string read_command_string(LLS* config) {
    auto x = read_command(config);
    string result = *(x->begin());
    delete x;
//    LOGD("Delete successful!\n");
    return result;
}

int read_command_integer(LLS* config) {
    string x = read_command_string(config);
//    LOGD("read_command_string successful!\n");
    return atoi(x.c_str());
}

long read_command_long(LLS* config) {
    string x = read_command_string(config);
    return atol(x.c_str());
}

double read_command_double(LLS* config) {
    string x = read_command_string(config);
    return atof(x.c_str());
}

long long read_command_long_long(LLS* config) {
    string x = read_command_string(config);
    return atoll(x.c_str());
}

void fatal_error(const char *msg) {
    perror(msg);
    exit(1);
}

int afs_open_fd(const char *path, int flags, LLS* lls) {
    send_command(lls, {OPEN_COMMAND, path, to_string(flags)});
    return read_command_long(lls);
}

int afs_close_fd(int fd, LLS* lls) {
    send_command(lls, {CLOSE_COMMAND, to_string(fd)});
    int ret = read_command_integer(lls);
    return -ret;
}

int afs_mkdir(const char *path, mode_t mode, LLS *lls) {
    send_command(lls, {MKDIR_COMMAND, path, to_string(mode)});
    int ret = read_command_integer(lls);
    return -ret;
}

int afs_truncate(const char *path, off_t offset, LLS *lls) {
    send_command(lls, {TRUNCATE_COMMAND, path, to_string(offset)});
    int ret = read_command_integer(lls);
    return -ret;
}

int afs_fallocate_fd(int fd, int mode, off_t offset, off_t len, LLS* lls) {
    send_command(lls, {FALLOCATE_COMMAND, to_string(fd), to_string(mode), to_string(offset), to_string(len)});
    int ret = read_command_integer(lls);
    return -ret;
}

int afs_read_fd(int fd, size_t size, off_t offset, char* buf, LLS* lls) {
    send_command(lls, {READ_COMMAND, to_string(fd), to_string(size), to_string(offset)});
    int read_bytes = read_command_integer(lls);
//    LOGI("Read bytes: %d\n", read_bytes);
    if (read_bytes != 0) {
        string read_base64 = read_command_string(lls);
//        LOGI("Read base64: %s\n", read_base64.c_str());
        unsigned char* read_decoded = base64_decode_bytes(read_base64);
        //strcpy(buf, read_decoded.c_str());
        memcpy(buf, read_decoded, read_bytes);
    } else {
        vector<string> *x = read_command(lls);
        buf[0] = 0;
        delete x;
    }
    return read_bytes;
}

int afs_write_fd(int fd, const char* buf, size_t size, off_t offset, LLS* lls) {
    string base64_encoded = base64_encode((const unsigned char*) buf, size);
    send_command(lls, {WRITE_COMMAND, to_string(fd), to_string(size), to_string(offset), base64_encoded});
    int ret = read_command_integer(lls);
//    LOGD("Returned integer: %d\n", ret);
    return ret;
}

int afs_create_fd(const char *path, mode_t mode, LLS* lls) {
    send_command(lls, {CREATE_COMMAND, path, to_string(mode)});
    return read_command_integer(lls);
}

int afs_unlink(const char *path, LLS *lls) {
    send_command(lls, {UNLINK_COMMAND, path});
    int ret = read_command_integer(lls);
    return -ret;
}

int afs_rmdir(const char *path, LLS *lls) {
    send_command(lls, {RMDIR_COMMAND, path});
    int ret = read_command_integer(lls);
    return -ret;
}

int afs_access(const char *path, int flags, LLS *lls) {
    send_command(lls, {ACCESS_COMMAND, path, to_string(flags)});
    int ret = read_command_integer(lls);
    return -ret;
}

int afs_futimens(const char *path, const struct timespec times[2], LLS* lls) {
    send_command(lls, {UTIMENS_COMMAND, path, to_string(times[0].tv_sec), to_string(times[0].tv_nsec), to_string(times[1].tv_sec), to_string(times[1].tv_nsec)});
    int ret = read_command_integer(lls);
    return -ret;
}

int afs_statfs(const char *path, struct statvfs *stbuf, LLS *lls) {
    send_command(lls, {STATFS_COMMAND, path});
    int retc = read_command_long_long(lls);
    long long _f_bsize = read_command_long_long(lls);
    long long _f_frsize = read_command_long_long(lls);
    long long _f_blocks = read_command_long_long(lls);
    long long _f_bfree = read_command_long_long(lls);
    long long _f_bavail = read_command_long_long(lls);
    long long _f_files = read_command_long_long(lls);
    long long _f_ffree = read_command_long_long(lls);
    long long _f_favail = read_command_long_long(lls);
    long long _f_fsid = read_command_long_long(lls);
    long long _f_flag = read_command_long_long(lls);
    long long _f_namemax = read_command_long_long(lls);
    if (retc == 0) {
        stbuf->f_bsize = _f_bsize;
        stbuf->f_frsize = _f_frsize;
        stbuf->f_blocks = _f_blocks;
        stbuf->f_bfree = _f_bfree;
        stbuf->f_bavail = _f_bavail;
        stbuf->f_files = _f_files;
        stbuf->f_ffree = _f_ffree;
        stbuf->f_favail = _f_favail;
        stbuf->f_fsid = _f_fsid;
        stbuf->f_flag = _f_flag;
        stbuf->f_namemax = _f_namemax;
    }
    return -retc;
}

int afs_getattr(const char *path, struct stat *stbuf, LLS *lls) {
    send_command(lls, {GETATTR_COMMAND, path});
    //memset(stbuf, 0, sizeof(struct stat));
    // We read struct stat by numbers in order of appearance in android source code.
    int retc = read_command_long_long(lls);
    long long st_dev = read_command_long_long(lls);
    long long st_ino = read_command_long_long(lls);
    long long st_mode = read_command_long_long(lls);
    long long st_nlink = read_command_long_long(lls);
    long long st_uid = read_command_long_long(lls);
    long long st_gid = read_command_long_long(lls);
    long long st_rdev = read_command_long_long(lls);
    long long st_size = read_command_long_long(lls);
    long long st_blksize = read_command_long_long(lls);
    long long st_blocks = read_command_long_long(lls);
    long long _st_atime = read_command_long_long(lls);
    long long _st_mtime = read_command_long_long(lls);
    long long _st_ctime = read_command_long_long(lls);
    if (retc == 0) {
        stbuf->st_dev = st_dev;
        stbuf->st_ino = st_ino;
        stbuf->st_mode = st_mode;
        stbuf->st_nlink = st_nlink;
        stbuf->st_uid = st_uid;
        stbuf->st_gid = st_gid;
        stbuf->st_rdev = st_rdev;
        stbuf->st_size = st_size;
        stbuf->st_blksize = st_blksize;
        stbuf->st_blocks = st_blocks;
        stbuf->st_atime = _st_atime;
        stbuf->st_mtime = _st_mtime;
        stbuf->st_ctime = _st_ctime;
    }
    return -retc;
}

int afs_rename(const char *oldpath, const char *newpath, LLS *lls) {
    send_command(lls, {RENAME_COMMAND, oldpath, newpath});
    int ret = read_command_integer(lls);
    return -ret;
}

string afs_identify_phone(LLS* lls) {
    send_command(lls, {IDENTIFY_COMMAND});
    return read_command_string(lls);
}

LLS *alc_connect(const char *ip_address, int port) {
    LLS* lls;
    lls = new LLS();
//    LOGI("Initiating connection\n");
    lls->protocol_version = new string("BETA1");
    lls->encrypted = false;
    lls->remote_port = port;
    lls->remote_ip_addr.sin_family = AF_INET;
    lls->remote_ip_addr.sin_port = htons(lls->remote_port);
    lls->comm_socket = socket(AF_INET, SOCK_STREAM, 0);
    if (lls->comm_socket <0) {
        delete lls;
        return NULL;
    }
    struct hostent *server = gethostbyname(ip_address);
    bcopy((char *)server->h_addr, (char *)&lls->remote_ip_addr.sin_addr.s_addr, server->h_length);
    if (connect(lls->comm_socket, (struct sockaddr *) &lls->remote_ip_addr, sizeof(lls->remote_ip_addr)) < 0) {
        delete lls;
        return NULL;
    }
//    LOGI("Connection established!\n");
    return lls;
}

void alc_disconnect(const LLS *lls) {
    close(lls->comm_socket);
}

void alc_pair(LLS *lls) {
    send_command(lls, {PAIR_COMMAND, alc_uuid()});
    // this command does not expect a reply!
}

bool alc_canencrypt(LLS *lls) {
    send_command(lls, {CANENCRYPT_COMMAND, alc_uuid()});
    string s = read_command_string(lls);
    string uuid = read_command_string(lls);
    return (s == RESPONSE_YES && has_key_for(uuid));
}

bool alc_startencrypt(LLS *lls) {
    send_command(lls, {ENCRYPT_COMMAND, alc_uuid(), DEFAULT_ENCRYPT_ALGORITHM});
    string s = read_command_string(lls);
    string android_uuid = read_command_string(lls);
    if (s == RESPONSE_OK) {
        alc_toggleencrypt(lls, android_uuid);
        return true;
    }
    return false;
}

void alc_toggleencrypt(LLS *lls, string uuid) {
//    LOGI("UUID: %s", uuid.c_str());
    string key = get_key(uuid);
//    LOGI("Base64 key: %s", key.c_str());
    unsigned char* key_bytes = base64_decode_bytes(key);
    memcpy(lls->encryption_key_bytes, key_bytes, 16);
    delete[] key_bytes;
    lls->encrypted = true;
}

bool alc_crypthello(LLS *lls) {
    string random_sequence = generate_uuid_string();
    send_command(lls, {CRYPTHELLO_COMMAND, random_sequence});
    string returned = read_command_string(lls);
    return returned == random_sequence;
}

bool afs_changeroot(LLS *lls, string newroot) {
    send_command(lls, {PREFIX_COMMAND, newroot});
    string returned = read_command_string(lls);
    return (returned == RESPONSE_OK);
}

vector<string> alc_get_supported_extensions(LLS *lls) {
    send_command(lls, {EXTENSIONS_COMMAND});
    int n = read_command_integer(lls);
    vector<string> result;
    while(n--) {
        result.push_back(read_command_string(lls));
    }
    return result;
}

void alc_enablesensors(LLS *lls) {
    send_command(lls, {STARTSENSOR_COMMAND});
}

void alc_disablesensors(LLS *lls) {
    send_command(lls, {QUITSENSOR_COMMAND});
}

vector<double> alc_getrotation(LLS *lls) {
    send_command(lls, {GETROTATION_COMMAND});
    double x = read_command_double(lls);
    double y = read_command_double(lls);
    double z = read_command_double(lls);
    return {x, y, z};
}